@@ -20,7 +20,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111(-1)307  USA
 {- |
    Module      : Text.Pandoc.Pretty
    Copyright   : Copyright (C) 2010 John MacFarlane
-   License     : GNU GPL, version 2 or above 
+   License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
    Stability   : alpha
@@ -202,18 +202,17 @@ chomp d = Doc (fromList dl')
 
 outp :: (IsString a, Monoid a)
      => Int -> String -> DocState a
-outp off s | off <= 0 = do
+outp off s | off < 0 = do  -- offset < 0 means newline characters
   st' <- get
   let rawpref = prefix st'
   when (column st' == 0 && usePrefix st' && not (null rawpref)) $ do
     let pref = reverse $ dropWhile isSpace $ reverse rawpref
     modify $ \st -> st{ output = fromString pref : output st
                       , column = column st + realLength pref }
-  when (off < 0) $ do
-     modify $ \st -> st { output = fromString s : output st
-                        , column = 0
-                        , newlines = newlines st + 1 }
-outp off s = do
+  modify $ \st -> st { output = fromString s : output st
+                     , column = 0
+                     , newlines = newlines st + 1 }
+outp off s = do           -- offset >= 0 (0 might be combining char)
   st' <- get
   let pref = prefix st'
   when (column st' == 0 && usePrefix st' && not (null pref)) $ do
@@ -510,7 +509,9 @@ charWidth c =
         | c >= '\xFE10' && c <= '\xFE19'   -> 2
         | c >= '\xFE20' && c <= '\xFE26'   -> 1
         | c >= '\xFE30' && c <= '\xFE6B'   -> 2
-        | c >= '\xFE70' && c <= '\x16A38'  -> 1
+        | c >= '\xFE70' && c <= '\xFEFF'   -> 1
+        | c >= '\xFF01' && c <= '\xFF60'   -> 2
+        | c >= '\xFF61' && c <= '\x16A38'  -> 1
         | c >= '\x1B000' && c <= '\x1B001' -> 2
         | c >= '\x1D000' && c <= '\x1F1FF' -> 1
         | c >= '\x1F200' && c <= '\x1F251' -> 2
