@@ -1,6 +1,6 @@
 module Text.Pandoc.Readers.DocBook ( readDocBook ) where
 import Data.Char (toUpper, isDigit)
-import Text.Pandoc.Parsing (ParserState(..))
+import Text.Pandoc.Options
 import Text.Pandoc.Definition
 import Text.Pandoc.Builder
 import Text.XML.Light
@@ -133,7 +133,7 @@ List of all DocBook tags, with [x] indicating implemented,
 [ ] exceptionname - The name of an exception
 [ ] fax - A fax number
 [ ] fieldsynopsis - The name of a field in a class definition
-[ ] figure - A formal figure, generally an illustration, with a title
+[x] figure - A formal figure, generally an illustration, with a title
 [x] filename - The name of a file
 [ ] firstname - The first name of a person
 [ ] firstterm - The first occurrence of a term
@@ -455,13 +455,13 @@ List of all DocBook tags, with [x] indicating implemented,
 [x] tocfront - An entry in a table of contents for a front matter component
 [x] toclevel1 - A top-level entry within a table of contents entry for a
     chapter-like component
-[x] toclevel2 - A second-level entry within a table of contents entry for a 
+[x] toclevel2 - A second-level entry within a table of contents entry for a
     chapter-like component
-[x] toclevel3 - A third-level entry within a table of contents entry for a 
+[x] toclevel3 - A third-level entry within a table of contents entry for a
     chapter-like component
-[x] toclevel4 - A fourth-level entry within a table of contents entry for a 
+[x] toclevel4 - A fourth-level entry within a table of contents entry for a
     chapter-like component
-[x] toclevel5 - A fifth-level entry within a table of contents entry for a 
+[x] toclevel5 - A fifth-level entry within a table of contents entry for a
     chapter-like component
 [x] tocpart - An entry in a table of contents for a part of a book
 [ ] token - A unit of information
@@ -501,9 +501,10 @@ data DBState = DBState{ dbSectionLevel :: Int
                       , dbDocAuthors   :: [Inlines]
                       , dbDocDate      :: Inlines
                       , dbBook         :: Bool
+                      , dbFigureTitle  :: Inlines
                       } deriving Show
 
-readDocBook :: ParserState -> String -> Pandoc
+readDocBook :: ReaderOptions -> String -> Pandoc
 readDocBook _ inp  = setTitle (dbDocTitle st')
                    $ setAuthors (dbDocAuthors st')
                    $ setDate (dbDocDate st')
@@ -515,8 +516,19 @@ readDocBook _ inp  = setTitle (dbDocTitle st')
                                     , dbDocAuthors = []
                                     , dbDocDate = mempty
                                     , dbBook = False
+                                    , dbFigureTitle = mempty
                                     }
 
+getFigure :: Element -> DB Blocks
+getFigure e = do
+  tit <- case filterChild (named "title") e of
+              Just t -> getInlines t
+              Nothing -> return mempty
+  modify $ \st -> st{ dbFigureTitle = tit }
+  res <- getBlocks e
+  modify $ \st -> st{ dbFigureTitle = mempty }
+  return res
+
 -- normalize input, consolidating adjacent Text and CRef elements
 normalizeTree :: [Content] -> [Content]
 normalizeTree = everywhere (mkT go)
@@ -574,7 +586,7 @@ addToStart toadd bs =
     (Para xs : rest) -> para (toadd <> fromList xs) <> fromList rest
     _                -> bs
 
--- function that is used by both mediaobject (in parseBlock) 
+-- function that is used by both mediaobject (in parseBlock)
 -- and inlinemediaobject (in parseInline)
 getImage :: Element -> DB Inlines
 getImage e = do
@@ -585,10 +597,13 @@ getImage e = do
                               Just i -> return $ attrValue "fileref" i
   caption <- case filterChild
                   (\x -> named "caption" x || named "textobject" x) e of
-               Nothing  -> return mempty
+               Nothing  -> gets dbFigureTitle
                Just z   -> mconcat <$> (mapM parseInline $ elContent z)
   return $ image imageUrl "" caption
 
+getBlocks :: Element -> DB Blocks
+getBlocks e =  mconcat <$> (mapM parseBlock $ elContent e)
+
 parseBlock :: Content -> DB Blocks
 parseBlock (Text (CData CDataRaw _ _)) = return mempty -- DOCTYPE
 parseBlock (Text (CData _ s _)) = if all isSpace s
@@ -613,7 +628,7 @@ parseBlock (Elem e) =
         "attribution" -> return mempty
         "titleabbrev" -> return mempty
         "authorinitials" -> return mempty
-        "title" -> return mempty -- handled by getTitle or sect
+        "title" -> return mempty -- handled by getTitle or sect or figure
         "bibliography" -> sect 0
         "bibliodiv" -> sect 1
         "biblioentry" -> parseMixed para (elContent e)
@@ -674,7 +689,8 @@ parseBlock (Elem e) =
           orderedListWith (start,listStyle,DefaultDelim)
             <$> listitems
         "variablelist" -> definitionList <$> deflistitems
-        "mediaobject" -> para <$> (getImage e)
+        "figure" -> getFigure e
+        "mediaobject" -> para <$> getImage e
         "caption" -> return mempty
         "info" -> getTitle >> getAuthors >> getDate >> return mempty
         "articleinfo" -> getTitle >> getAuthors >> getDate >> return mempty
@@ -702,8 +718,7 @@ parseBlock (Elem e) =
         "programlisting" -> codeBlockWithLang
         "?xml"  -> return mempty
         _       -> getBlocks e
-   where getBlocks e' =  mconcat <$> (mapM parseBlock $ elContent e')
-         parseMixed container conts = do
+   where parseMixed container conts = do
            let (ils,rest) = break isBlockElement conts
            ils' <- (trimInlines . mconcat) <$> mapM parseInline ils
            let p = if ils' == mempty then mempty else container ils'
@@ -862,15 +877,15 @@ parseInline (Elem e) =
         "varargs" -> return $ code "(...)"
         "xref" -> return $ str "?" -- so at least you know something is there
         "email" -> return $ link ("mailto:" ++ strContent e) ""
-                          $ code $ strContent e
-        "uri" -> return $ link (strContent e) "" $ code $ strContent e
+                          $ str $ strContent e
+        "uri" -> return $ link (strContent e) "" $ str $ strContent e
         "ulink" -> link (attrValue "url" e) "" <$> innerInlines
         "link" -> do
              ils <- innerInlines
              let href = case findAttr (QName "href" (Just "http://www.w3.org/1999/xlink") Nothing) e of
                                Just h -> h
                                _      -> ('#' : attrValue "linkend" e)
-             let ils' = if ils == mempty then code href else ils
+             let ils' = if ils == mempty then str href else ils
              return $ link href "" ils'
         "foreignphrase" -> emph <$> innerInlines
         "emphasis" -> case attrValue "role" e of
