@@ -20,10 +20,10 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 {- |
    Module      : Text.Pandoc.Writers.ConTeXt
    Copyright   : Copyright (C) 2007-2010 John MacFarlane
-   License     : GNU GPL, version 2 or above 
+   License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
-   Stability   : alpha 
+   Stability   : alpha
    Portability : portable
 
 Conversion of 'Pandoc' format into ConTeXt.
@@ -31,31 +31,32 @@ Conversion of 'Pandoc' format into ConTeXt.
 module Text.Pandoc.Writers.ConTeXt ( writeConTeXt ) where
 import Text.Pandoc.Definition
 import Text.Pandoc.Shared
+import Text.Pandoc.Options
 import Text.Pandoc.Generic (queryWith)
 import Text.Printf ( printf )
-import Data.List ( intercalate )
+import Data.List ( intercalate, isPrefixOf )
 import Control.Monad.State
 import Text.Pandoc.Pretty
 import Text.Pandoc.Templates ( renderTemplate )
 import Network.URI ( isURI, unEscapeString )
 
-data WriterState = 
+data WriterState =
   WriterState { stNextRef          :: Int  -- number of next URL reference
               , stOrderedListLevel :: Int  -- level of ordered list
               , stOptions          :: WriterOptions -- writer options
               }
 
-orderedListStyles :: [[Char]]
-orderedListStyles = cycle ["[n]","[a]", "[r]", "[g]"] 
+orderedListStyles :: [Char]
+orderedListStyles = cycle "narg"
 
 -- | Convert Pandoc to ConTeXt.
 writeConTeXt :: WriterOptions -> Pandoc -> String
-writeConTeXt options document = 
+writeConTeXt options document =
   let defaultWriterState = WriterState { stNextRef = 1
                                        , stOrderedListLevel = 0
                                        , stOptions = options
-                                       } 
-  in evalState (pandocToConTeXt options document) defaultWriterState 
+                                       }
+  in evalState (pandocToConTeXt options document) defaultWriterState
 
 pandocToConTeXt :: WriterOptions -> Pandoc -> State WriterState String
 pandocToConTeXt options (Pandoc (Meta title authors date) blocks) = do
@@ -73,6 +74,12 @@ pandocToConTeXt options (Pandoc (Meta title authors date) blocks) = do
   let main = (render colwidth . vcat) body
   let context  = writerVariables options ++
                  [ ("toc", if writerTableOfContents options then "yes" else "")
+                 , ("placelist", intercalate "," $
+                     take (writerTOCDepth options + if writerChapters options
+                                                       then 0
+                                                       else 1)
+                       ["chapter","section","subsection","subsubsection",
+                        "subsubsubsection","subsubsubsubsection"])
                  , ("body", main)
                  , ("title", titletext)
                  , ("date", datetext) ] ++
@@ -120,15 +127,16 @@ elementToConTeXt opts (Sec level _ id' title' elements) = do
   return $ vcat (header' : innerContents)
 
 -- | Convert Pandoc block element to ConTeXt.
-blockToConTeXt :: Block 
+blockToConTeXt :: Block
                -> State WriterState Doc
 blockToConTeXt Null = return empty
 blockToConTeXt (Plain lst) = inlineListToConTeXt lst
-blockToConTeXt (Para [Image txt (src,_)]) = do
+-- title beginning with fig: indicates that the image is a figure
+blockToConTeXt (Para [Image txt (src,'f':'i':'g':':':_)]) = do
   capt <- inlineListToConTeXt txt
   return $ blankline $$ "\\placefigure[here,nonumber]" <> braces capt <>
            braces ("\\externalfigure" <> brackets (text src)) <> blankline
-blockToConTeXt (Para lst) = do 
+blockToConTeXt (Para lst) = do
   contents <- inlineListToConTeXt lst
   return $ contents <> blankline
 blockToConTeXt (BlockQuote lst) = do
@@ -141,37 +149,41 @@ blockToConTeXt (RawBlock "context" str) = return $ text str <> blankline
 blockToConTeXt (RawBlock _ _ ) = return empty
 blockToConTeXt (BulletList lst) = do
   contents <- mapM listItemToConTeXt lst
-  return $ "\\startitemize" $$ vcat contents $$ text "\\stopitemize" <> blankline
+  return $ ("\\startitemize" <> if isTightList lst
+                                   then brackets "packed"
+                                   else empty) $$
+    vcat contents $$ text "\\stopitemize" <> blankline
 blockToConTeXt (OrderedList (start, style', delim) lst) = do
     st <- get
     let level = stOrderedListLevel st
     put $ st {stOrderedListLevel = level + 1}
     contents <- mapM listItemToConTeXt lst
-    put $ st {stOrderedListLevel = level} 
+    put $ st {stOrderedListLevel = level}
     let start' = if start == 1 then "" else "start=" ++ show start
     let delim' = case delim of
                         DefaultDelim -> ""
-                        Period       -> "stopper=." 
-                        OneParen     -> "stopper=)" 
+                        Period       -> "stopper=."
+                        OneParen     -> "stopper=)"
                         TwoParens    -> "left=(,stopper=)"
-    let width = maximum $ map length $ take (length contents) 
+    let width = maximum $ map length $ take (length contents)
                           (orderedListMarkers (start, style', delim))
     let width' = (toEnum width + 1) / 2
-    let width'' = if width' > (1.5 :: Double) 
-                     then "width=" ++ show width' ++ "em" 
+    let width'' = if width' > (1.5 :: Double)
+                     then "width=" ++ show width' ++ "em"
                      else ""
     let specs2Items = filter (not . null) [start', delim', width'']
     let specs2 = if null specs2Items
                     then ""
                     else "[" ++ intercalate "," specs2Items ++ "]"
-    let style'' = case style' of
-                        DefaultStyle -> orderedListStyles !! level
-                        Decimal      -> "[n]" 
-                        Example      -> "[n]" 
-                        LowerRoman   -> "[r]"
-                        UpperRoman   -> "[R]"
-                        LowerAlpha   -> "[a]"
-                        UpperAlpha   -> "[A]"
+    let style'' = '[': (case style' of
+                          DefaultStyle -> orderedListStyles !! level
+                          Decimal      -> 'n'
+                          Example      -> 'n'
+                          LowerRoman   -> 'r'
+                          UpperRoman   -> 'R'
+                          LowerAlpha   -> 'a'
+                          UpperAlpha   -> 'A') :
+                       if isTightList lst then ",packed]" else "]"
     let specs = style'' ++ specs2
     return $ "\\startitemize" <> text specs $$ vcat contents $$
              "\\stopitemize" <> blankline
@@ -179,24 +191,24 @@ blockToConTeXt (DefinitionList lst) =
   liftM vcat $ mapM defListItemToConTeXt lst
 blockToConTeXt HorizontalRule = return $ "\\thinrule" <> blankline
 -- If this is ever executed, provide a default for the reference identifier.
-blockToConTeXt (Header level lst) = sectionHeader "" level lst
+blockToConTeXt (Header level (ident,_,_) lst) = sectionHeader ident level lst
 blockToConTeXt (Table caption aligns widths heads rows) = do
     let colDescriptor colWidth alignment = (case alignment of
-                                               AlignLeft    -> 'l' 
+                                               AlignLeft    -> 'l'
                                                AlignRight   -> 'r'
                                                AlignCenter  -> 'c'
                                                AlignDefault -> 'l'):
            if colWidth == 0
               then "|"
               else ("p(" ++ printf "%.2f" colWidth ++ "\\textwidth)|")
-    let colDescriptors = "|" ++ (concat $ 
+    let colDescriptors = "|" ++ (concat $
                                  zipWith colDescriptor widths aligns)
     headers <- if all null heads
                   then return empty
-                  else liftM ($$ "\\HL") $ tableRowToConTeXt heads 
-    captionText <- inlineListToConTeXt caption 
+                  else liftM ($$ "\\HL") $ tableRowToConTeXt heads
+    captionText <- inlineListToConTeXt caption
     let captionText' = if null caption then text "none" else captionText
-    rows' <- mapM tableRowToConTeXt rows 
+    rows' <- mapM tableRowToConTeXt rows
     return $ "\\placetable[here]" <> braces captionText' $$
              "\\starttable" <> brackets (text colDescriptors) $$
              "\\HL" $$ headers $$
@@ -230,7 +242,7 @@ inlineListToConTeXt lst = liftM hcat $ mapM inlineToConTeXt lst
 -- | Convert inline element to ConTeXt
 inlineToConTeXt :: Inline    -- ^ Inline to convert
                 -> State WriterState Doc
-inlineToConTeXt (Emph lst) = do 
+inlineToConTeXt (Emph lst) = do
   contents <- inlineListToConTeXt lst
   return $ braces $ "\\em " <> contents
 inlineToConTeXt (Strong lst) = do
@@ -273,7 +285,11 @@ inlineToConTeXt (RawInline _ _) = return empty
 inlineToConTeXt (LineBreak) = return $ text "\\crlf" <> cr
 inlineToConTeXt Space = return space
 -- autolink
-inlineToConTeXt (Link [Code _ str] (src, tit)) = inlineToConTeXt (Link
+inlineToConTeXt (Link [Str str] (src, tit))
+  | if "mailto:" `isPrefixOf` src
+    then src == escapeURI ("mailto:" ++ str)
+    else src == escapeURI str =
+  inlineToConTeXt (Link
     [RawInline "context" "\\hyphenatedurl{", Str str, RawInline "context" "}"]
     (src, tit))
 -- Handle HTML-like internal document references to sections
