@@ -1,3 +1,4 @@
+{-# LANGUAGE ScopedTypeVariables #-}
 {-
 Copyright (C) 2012 John MacFarlane <jgm@berkeley.edu>
 
@@ -29,20 +30,17 @@ Conversion of 'Pandoc' documents to docx.
 -}
 module Text.Pandoc.Writers.Docx ( writeDocx ) where
 import Data.List ( intercalate )
-import System.FilePath ( (</>) )
-import qualified Data.ByteString.Lazy as B
+import qualified Data.ByteString as B
+import qualified Data.ByteString.Lazy as BL
 import qualified Data.Map as M
-import Data.ByteString.Lazy.UTF8 ( fromString, toString )
-import Text.Pandoc.UTF8 as UTF8
-import System.IO ( stderr )
+import qualified Text.Pandoc.UTF8 as UTF8
 import Codec.Archive.Zip
 import Data.Time.Clock.POSIX
-import Paths_pandoc ( getDataFileName )
 import Text.Pandoc.Definition
 import Text.Pandoc.Generic
-import System.Directory
 import Text.Pandoc.ImageSize
 import Text.Pandoc.Shared hiding (Element)
+import Text.Pandoc.Options
 import Text.Pandoc.Readers.TeXMath
 import Text.Pandoc.Highlighting ( highlight )
 import Text.Highlighting.Kate.Types ()
@@ -50,6 +48,10 @@ import Text.XML.Light
 import Text.TeXMath
 import Control.Monad.State
 import Text.Highlighting.Kate
+import Data.Unique (hashUnique, newUnique)
+import System.Random (randomRIO)
+import Text.Printf (printf)
+import qualified Control.Exception as E
 
 data WriterState = WriterState{
          stTextProperties :: [Element]
@@ -57,9 +59,9 @@ data WriterState = WriterState{
        , stFootnotes      :: [Element]
        , stSectionIds     :: [String]
        , stExternalLinks  :: M.Map String String
-       , stImages         :: M.Map FilePath (String, B.ByteString)
+       , stImages         :: M.Map FilePath (String, String, Element, B.ByteString)
        , stListLevel      :: Int
-       , stListMarker     :: ListMarker
+       , stListNumId      :: Int
        , stNumStyles      :: M.Map ListMarker Int
        , stLists          :: [ListMarker]
        }
@@ -78,7 +80,7 @@ defaultWriterState = WriterState{
       , stExternalLinks  = M.empty
       , stImages         = M.empty
       , stListLevel      = -1
-      , stListMarker     = NoMarker
+      , stListNumId      = 1
       , stNumStyles      = M.fromList [(NoMarker, 0)]
       , stLists          = [NoMarker]
       }
@@ -92,68 +94,62 @@ mknode :: Node t => String -> [(String,String)] -> t -> Element
 mknode s attrs =
   add_attrs (map (\(k,v) -> Attr (unqual k) v) attrs) . node (unqual s)
 
--- | Produce an Docx file from a Pandoc document.
-writeDocx :: Maybe FilePath -- ^ Path specified by --reference-docx
-          -> WriterOptions  -- ^ Writer options
-          -> Pandoc         -- ^ Document to convert
-          -> IO B.ByteString
-writeDocx mbRefDocx opts doc@(Pandoc (Meta tit auths date) _) = do
-  let datadir = writerUserDataDir opts
-  refArchive <- liftM toArchive $
-       case mbRefDocx of
-             Just f -> B.readFile f
-             Nothing -> do
-               let defaultDocx = getDataFileName "reference.docx" >>= B.readFile
-               case datadir of
-                     Nothing  -> defaultDocx
-                     Just d   -> do
-                        exists <- doesFileExist (d </> "reference.docx")
-                        if exists
-                           then B.readFile (d </> "reference.docx")
-                           else defaultDocx
+toLazy :: B.ByteString -> BL.ByteString
+toLazy = BL.fromChunks . (:[])
 
-  (newContents, st) <- runStateT (writeOpenXML opts{writerWrapText = False} doc)
+-- | Produce an Docx file from a Pandoc document.
+writeDocx :: WriterOptions  -- ^ Writer options
+          -> Pandoc         -- ^ Document to convert
+          -> IO BL.ByteString
+writeDocx opts doc@(Pandoc (Meta tit auths date) _) = do
+  let datadir = writerUserDataDir opts
+  refArchive <- liftM (toArchive . toLazy) $
+       case writerReferenceDocx opts of
+             Just f  -> B.readFile f
+             Nothing -> readDataFile datadir "reference.docx"
+
+  ((contents, footnotes), st) <- runStateT (writeOpenXML opts{writerWrapText = False} doc)
                        defaultWriterState
   epochtime <- floor `fmap` getPOSIXTime
   let imgs = M.elems $ stImages st
-  let imgPath ident img = "media/" ++ ident ++
-                            case imageType img of
-                                  Just Png  -> ".png"
-                                  Just Jpeg -> ".jpeg"
-                                  Just Gif  -> ".gif"
-                                  Nothing   -> ""
-  let toImgRel (ident,img) =  mknode "Relationship" [("Type","http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"),("Id",ident),("Target",imgPath ident img)] ()
+  let toImgRel (ident,path,_,_) =  mknode "Relationship" [("Type","http://schemas.openxmlformats.org/officeDocument/2006/relationships/image"),("Id",ident),("Target",path)] ()
   let newrels = map toImgRel imgs
   let relpath = "word/_rels/document.xml.rels"
   let reldoc = case findEntryByPath relpath refArchive >>=
-                    parseXMLDoc . toString . fromEntry of
+                    parseXMLDoc . UTF8.toStringLazy . fromEntry of
                       Just d  -> d
                       Nothing -> error $ relpath ++ "missing in reference docx"
   let reldoc' = reldoc{ elContent = elContent reldoc ++ map Elem newrels }
   -- create entries for images
-  let toImageEntry (ident,img) = toEntry ("word/" ++ imgPath ident img)
-         epochtime img
+  let toImageEntry (_,path,_,img) = toEntry ("word/" ++ path) epochtime $ toLazy img
   let imageEntries = map toImageEntry imgs
   -- NOW get list of external links and images from this, and do what's needed
   let toLinkRel (src,ident) =  mknode "Relationship" [("Type","http://schemas.openxmlformats.org/officeDocument/2006/relationships/hyperlink"),("Id",ident),("Target",src),("TargetMode","External") ] ()
   let newrels' = map toLinkRel $ M.toList $ stExternalLinks st
   let reldoc'' = reldoc' { elContent = elContent reldoc' ++ map Elem newrels' }
-  let relEntry = toEntry relpath epochtime $ fromString $ showTopElement' reldoc''
-  let contentEntry = toEntry "word/document.xml" epochtime $ fromString $ showTopElement' newContents
+  let relEntry = toEntry relpath epochtime $ UTF8.fromStringLazy $ showTopElement' reldoc''
+  let contentEntry = toEntry "word/document.xml" epochtime $ UTF8.fromStringLazy $ showTopElement' contents
+  -- footnotes
+  let footnotesEntry = toEntry "word/footnotes.xml" epochtime $ UTF8.fromStringLazy $
+        showTopElement' footnotes
+  -- footnote rels
+  let footnoteRelEntry = toEntry "word/_rels/footnotes.xml.rels" epochtime $ UTF8.fromStringLazy $
+        showTopElement' $ mknode "Relationships" [("xmlns","http://schemas.openxmlformats.org/package/2006/relationships")]
+        $ newrels'
   -- styles
   let newstyles = styleToOpenXml $ writerHighlightStyle opts
   let stylepath = "word/styles.xml"
   let styledoc = case findEntryByPath stylepath refArchive >>=
-                      parseXMLDoc . toString . fromEntry of
+                      parseXMLDoc . UTF8.toStringLazy . fromEntry of
                         Just d  -> d
                         Nothing -> error $ "Unable to parse " ++ stylepath ++
                                            " from reference.docx"
   let styledoc' = styledoc{ elContent = elContent styledoc ++ map Elem newstyles }
-  let styleEntry = toEntry stylepath epochtime $ fromString $ showTopElement' styledoc'
+  let styleEntry = toEntry stylepath epochtime $ UTF8.fromStringLazy $ showTopElement' styledoc'
   -- construct word/numbering.xml
   let numpath = "word/numbering.xml"
-  let numEntry = toEntry numpath epochtime $ fromString $ showTopElement'
-                 $ mkNumbering (stNumStyles st) (stLists st)
+  numEntry <- (toEntry numpath epochtime . UTF8.fromStringLazy . showTopElement')
+                 `fmap` mkNumbering (stNumStyles st) (stLists st)
   let docPropsPath = "docProps/core.xml"
   let docProps = mknode "cp:coreProperties"
           [("xmlns:cp","http://schemas.openxmlformats.org/package/2006/metadata/core-properties")
@@ -166,18 +162,18 @@ writeDocx mbRefDocx opts doc@(Pandoc (Meta tit auths date) _) = do
             (maybe "" id $ normalizeDate $ stringify date)
           : mknode "dcterms:modified" [("xsi:type","dcterms:W3CDTF")] () -- put current time here
           : map (mknode "dc:creator" [] . stringify) auths
-  let docPropsEntry = toEntry docPropsPath epochtime $ fromString $ showTopElement' docProps
+  let docPropsEntry = toEntry docPropsPath epochtime $ UTF8.fromStringLazy $ showTopElement' docProps
   let relsPath = "_rels/.rels"
   rels <- case findEntryByPath relsPath refArchive of
-                   Just e  -> return $ toString $ fromEntry e
+                   Just e  -> return $ UTF8.toStringLazy $ fromEntry e
                    Nothing -> err 57 "could not find .rels/_rels in reference docx"
   -- fix .rels/_rels, which can get screwed up when reference.docx is edited by Word
   let rels' = substitute "http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties"
                   "http://schemas.openxmlformats.org/officedocument/2006/relationships/metadata/core-properties"
                   rels
-  let relsEntry = toEntry relsPath epochtime $ fromString rels'
+  let relsEntry = toEntry relsPath epochtime $ UTF8.fromStringLazy rels'
   let archive = foldr addEntryToArchive refArchive $
-                  relsEntry : contentEntry : relEntry : numEntry : styleEntry : docPropsEntry : imageEntries
+                  relsEntry : contentEntry : relEntry : footnoteRelEntry : numEntry : styleEntry : footnotesEntry : docPropsEntry : imageEntries
   return $ fromArchive archive
 
 styleToOpenXml :: Style -> [Element]
@@ -215,11 +211,12 @@ styleToOpenXml style = parStyle : map toStyle alltoktypes
                                  $ backgroundColor style )
                              ]
 
-mkNumbering :: M.Map ListMarker Int -> [ListMarker] -> Element
-mkNumbering markers lists =
-  mknode "w:numbering" [("xmlns:w","http://schemas.openxmlformats.org/wordprocessingml/2006/main")]
-   $  map mkAbstractNum (M.toList markers)
-   ++ zipWith (mkNum markers) lists [1..(length lists)]
+mkNumbering :: M.Map ListMarker Int -> [ListMarker] -> IO Element
+mkNumbering markers lists = do
+  elts <- mapM mkAbstractNum (M.toList markers)
+  return $ mknode "w:numbering"
+     [("xmlns:w","http://schemas.openxmlformats.org/wordprocessingml/2006/main")]
+     $ elts ++ zipWith (mkNum markers) lists [1..(length lists)]
 
 mkNum :: M.Map ListMarker Int -> ListMarker -> Int -> Element
 mkNum markers marker numid =
@@ -233,10 +230,12 @@ mkNum markers marker numid =
               $ mknode "w:startOverride" [("w:val",show start)] ()) [0..6]
    where absnumid = maybe 0 id $ M.lookup marker markers
 
-mkAbstractNum :: (ListMarker,Int) -> Element
-mkAbstractNum (marker,numid) =
-  mknode "w:abstractNum" [("w:abstractNumId",show numid)]
-    $ mknode "w:multiLevelType" [("w:val","multilevel")] ()
+mkAbstractNum :: (ListMarker,Int) -> IO Element
+mkAbstractNum (marker,numid) = do
+  nsid <- randomRIO (0x10000000 :: Integer, 0xFFFFFFFF :: Integer)
+  return $ mknode "w:abstractNum" [("w:abstractNumId",show numid)]
+    $ mknode "w:nsid" [("w:val", printf "%8x" nsid)] ()
+    : mknode "w:multiLevelType" [("w:val","multilevel")] ()
     : map (mkLvl marker) [0..6]
 
 mkLvl :: ListMarker -> Int -> Element
@@ -285,8 +284,11 @@ mkLvl marker lvl =
           patternFor TwoParens s = "(" ++ s ++ ")"
           patternFor _ s = s ++ "."
 
--- | Convert Pandoc document to string in OpenXML format.
-writeOpenXML :: WriterOptions -> Pandoc -> WS Element
+getNumId :: WS Int
+getNumId = length `fmap` gets stLists
+
+-- | Convert Pandoc document to two OpenXML elements (the main document and footnotes).
+writeOpenXML :: WriterOptions -> Pandoc -> WS (Element, Element)
 writeOpenXML opts (Pandoc (Meta tit auths dat) blocks) = do
   title <- withParaProp (pStyle "Title") $ blocksToOpenXML opts [Para tit | not (null tit)]
   authors <- withParaProp (pStyle "Authors") $ blocksToOpenXML opts
@@ -296,13 +298,10 @@ writeOpenXML opts (Pandoc (Meta tit auths dat) blocks) = do
       convertSpace (Str x : Str y : xs) = Str (x ++ y) : xs
       convertSpace xs = xs
   let blocks' = bottomUp convertSpace $ blocks
-  doc <- blocksToOpenXML opts blocks'
+  doc' <- blocksToOpenXML opts blocks'
   notes' <- reverse `fmap` gets stFootnotes
-  let notes = case notes' of
-                   [] -> []
-                   ns -> [mknode "w:footnotes" [] ns]
   let meta = title ++ authors ++ date
-  return $ mknode "w:document"
+  let stdAttributes =
             [("xmlns:w","http://schemas.openxmlformats.org/wordprocessingml/2006/main")
             ,("xmlns:m","http://schemas.openxmlformats.org/officeDocument/2006/math")
             ,("xmlns:r","http://schemas.openxmlformats.org/officeDocument/2006/relationships")
@@ -312,7 +311,9 @@ writeOpenXML opts (Pandoc (Meta tit auths dat) blocks) = do
             ,("xmlns:a","http://schemas.openxmlformats.org/drawingml/2006/main")
             ,("xmlns:pic","http://schemas.openxmlformats.org/drawingml/2006/picture")
             ,("xmlns:wp","http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing")]
-         $ mknode "w:body" [] (meta ++ doc ++ notes)
+  let doc = mknode "w:document" stdAttributes $ mknode "w:body" [] (meta ++ doc')
+  let notes = mknode "w:footnotes" stdAttributes notes'
+  return (doc, notes)
 
 -- | Convert a list of Pandoc blocks to OpenXML.
 blocksToOpenXML :: WriterOptions -> [Block] -> WS [Element]
@@ -324,23 +325,32 @@ pStyle sty = mknode "w:pStyle" [("w:val",sty)] ()
 rStyle :: String -> Element
 rStyle sty = mknode "w:rStyle" [("w:val",sty)] ()
 
+getUniqueId :: MonadIO m => m String
+-- the + 20 is to ensure that there are no clashes with the rIds
+-- already in word/document.xml.rel
+getUniqueId = liftIO $ (show . (+ 20) . hashUnique) `fmap` newUnique
+
 -- | Convert a Pandoc block element to OpenXML.
 blockToOpenXML :: WriterOptions -> Block -> WS [Element]
 blockToOpenXML _ Null = return []
-blockToOpenXML opts (Header lev lst) = do
+blockToOpenXML opts (Header lev (ident,_,_) lst) = do
   contents <- withParaProp (pStyle $ "Heading" ++ show lev) $
                blockToOpenXML opts (Para lst)
   usedIdents <- gets stSectionIds
-  let ident = uniqueIdent lst usedIdents
-  modify $ \s -> s{ stSectionIds = ident : stSectionIds s }
-  let bookmarkStart = mknode "w:bookmarkStart" [("w:id",ident)
-                                               ,("w:name",ident)] ()
-  let bookmarkEnd = mknode "w:bookmarkEnd" [("w:id",ident)] ()
+  let bookmarkName = if null ident
+                        then uniqueIdent lst usedIdents
+                        else ident
+  modify $ \s -> s{ stSectionIds = bookmarkName : stSectionIds s }
+  id' <- getUniqueId
+  let bookmarkStart = mknode "w:bookmarkStart" [("w:id", id')
+                                               ,("w:name",bookmarkName)] ()
+  let bookmarkEnd = mknode "w:bookmarkEnd" [("w:id", id')] ()
   return $ [bookmarkStart] ++ contents ++ [bookmarkEnd]
 blockToOpenXML opts (Plain lst) = blockToOpenXML opts (Para lst)
-blockToOpenXML opts (Para x@[Image alt _]) = do
+-- title beginning with fig: indicates that the image is a figure
+blockToOpenXML opts (Para [Image alt (src,'f':'i':'g':':':tit)]) = do
   paraProps <- getParaProps
-  contents <- inlinesToOpenXML opts x
+  contents <- inlinesToOpenXML opts [Image alt (src,tit)]
   captionNode <- withParaProp (pStyle "ImageCaption")
                  $ blockToOpenXML opts (Para alt)
   return $ mknode "w:p" [] (paraProps ++ contents) : captionNode
@@ -402,11 +412,13 @@ blockToOpenXML opts (Table caption aligns widths headers rows) = do
 blockToOpenXML opts (BulletList lst) = do
   let marker = BulletMarker
   addList marker
-  asList $ concat `fmap` mapM (listItemToOpenXML opts marker) lst
+  numid  <- getNumId
+  asList $ concat `fmap` mapM (listItemToOpenXML opts numid) lst
 blockToOpenXML opts (OrderedList (start, numstyle, numdelim) lst) = do
   let marker = NumberMarker numstyle numdelim start
   addList marker
-  asList $ concat `fmap` mapM (listItemToOpenXML opts marker) lst
+  numid  <- getNumId
+  asList $ concat `fmap` mapM (listItemToOpenXML opts numid) lst
 blockToOpenXML opts (DefinitionList items) =
   concat `fmap` mapM (definitionListItemToOpenXML opts) items
 
@@ -418,9 +430,6 @@ definitionListItemToOpenXML opts (term,defs) = do
            $ concat `fmap` mapM (blocksToOpenXML opts) defs
   return $ term' ++ defs'
 
-getNumId :: WS Int
-getNumId = length `fmap` gets stLists
-
 addList :: ListMarker -> WS ()
 addList marker = do
   lists <- gets stLists
@@ -431,11 +440,11 @@ addList marker = do
            Nothing -> modify $ \st ->
                  st{ stNumStyles = M.insert marker (M.size numStyles + 1) numStyles }
 
-listItemToOpenXML :: WriterOptions -> ListMarker -> [Block] -> WS [Element]
+listItemToOpenXML :: WriterOptions -> Int -> [Block] -> WS [Element]
 listItemToOpenXML _ _ []                   = return []
-listItemToOpenXML opts marker (first:rest) = do
-  first' <- withMarker marker $ blockToOpenXML opts first
-  rest'  <- withMarker NoMarker $ blocksToOpenXML opts rest
+listItemToOpenXML opts numid (first:rest) = do
+  first' <- withNumId numid $ blockToOpenXML opts first
+  rest'  <- withNumId 1      $ blocksToOpenXML opts rest
   return $ first' ++ rest'
 
 alignmentToString :: Alignment -> [Char]
@@ -449,12 +458,12 @@ alignmentToString alignment = case alignment of
 inlinesToOpenXML :: WriterOptions -> [Inline] -> WS [Element]
 inlinesToOpenXML opts lst = concat `fmap` mapM (inlineToOpenXML opts) lst
 
-withMarker :: ListMarker -> WS a -> WS a
-withMarker m p = do
-  origMarker <- gets stListMarker
-  modify $ \st -> st{ stListMarker = m }
+withNumId :: Int -> WS a -> WS a
+withNumId numid p = do
+  origNumId <- gets stListNumId
+  modify $ \st -> st{ stListNumId = numid }
   result <- p
-  modify $ \st -> st{ stListMarker = origMarker }
+  modify $ \st -> st{ stListNumId = origNumId }
   return result
 
 asList :: WS a -> WS a
@@ -489,10 +498,7 @@ getParaProps :: WS [Element]
 getParaProps = do
   props <- gets stParaProperties
   listLevel <- gets stListLevel
-  listMarker <- gets stListMarker
-  numid <- case listMarker of
-                 NoMarker -> return 1
-                 _        -> getNumId
+  numid <- gets stListNumId
   let listPr = if listLevel >= 0
                   then [ mknode "w:numPr" []
                          [ mknode "w:numId" [("w:val",show numid)] ()
@@ -543,7 +549,7 @@ inlineToOpenXML opts (SmallCaps lst) =
 inlineToOpenXML opts (Strikeout lst) =
   withTextProp (mknode "w:strike" [] ())
   $ inlinesToOpenXML opts lst
-inlineToOpenXML _ LineBreak = return [ mknode "w:br" [] () ]
+inlineToOpenXML _ LineBreak = return [br]
 inlineToOpenXML _ (RawInline f str)
   | f == "openxml" = return [ x | Elem x <- parseXML str ]
   | otherwise      = return []
@@ -562,23 +568,21 @@ inlineToOpenXML opts (Math DisplayMath str) =
         Left  _ -> do
             fallback <- inlinesToOpenXML opts (readTeXMath str)
             return $ [br] ++ fallback ++ [br]
-    where br = mknode "w:br" [] ()
 inlineToOpenXML opts (Cite _ lst) = inlinesToOpenXML opts lst
 inlineToOpenXML _ (Code attrs str) =
   withTextProp (rStyle "VerbatimChar")
   $ case highlight formatOpenXML attrs str of
-         Nothing  -> intercalate [mknode "w:br" [] ()]
+         Nothing  -> intercalate [br]
                      `fmap` (mapM formattedString $ lines str)
          Just h   -> return h
-     where formatOpenXML _fmtOpts = intercalate [mknode "w:br" [] ()] .
-                                    map (map toHlTok)
+     where formatOpenXML _fmtOpts = intercalate [br] . map (map toHlTok)
            toHlTok (toktype,tok) = mknode "w:r" []
                                      [ mknode "w:rPr" []
                                        [ rStyle $ show toktype ]
                                      , mknode "w:t" [("xml:space","preserve")] tok ]
 inlineToOpenXML opts (Note bs) = do
   notes <- gets stFootnotes
-  let notenum = length notes + 1
+  notenum <- getUniqueId
   let notemarker = mknode "w:r" []
                    [ mknode "w:rPr" [] (rStyle "FootnoteReference")
                    , mknode "w:footnoteRef" [] () ]
@@ -594,11 +598,11 @@ inlineToOpenXML opts (Note bs) = do
                 $ insertNoteRef bs
   modify $ \st -> st{ stListLevel = oldListLevel, stParaProperties = oldParaProperties,
                       stTextProperties = oldTextProperties }
-  let newnote = mknode "w:footnote" [("w:id",show notenum)] $ contents
+  let newnote = mknode "w:footnote" [("w:id", notenum)] $ contents
   modify $ \s -> s{ stFootnotes = newnote : notes }
   return [ mknode "w:r" []
            [ mknode "w:rPr" [] (rStyle "FootnoteReference")
-           , mknode "w:footnoteReference" [("w:id", show notenum)] () ] ]
+           , mknode "w:footnoteReference" [("w:id", notenum)] () ] ]
 -- internal link:
 inlineToOpenXML opts (Link txt ('#':xs,_)) = do
   contents <- withTextProp (rStyle "Hyperlink") $ inlinesToOpenXML opts txt
@@ -607,65 +611,76 @@ inlineToOpenXML opts (Link txt ('#':xs,_)) = do
 inlineToOpenXML opts (Link txt (src,_)) = do
   contents <- withTextProp (rStyle "Hyperlink") $ inlinesToOpenXML opts txt
   extlinks <- gets stExternalLinks
-  ind <- case M.lookup src extlinks of
+  id' <- case M.lookup src extlinks of
             Just i   -> return i
             Nothing  -> do
-              let i = "link" ++ show (M.size extlinks)
+              i <- ("rId"++) `fmap` getUniqueId
               modify $ \st -> st{ stExternalLinks =
                         M.insert src i extlinks }
               return i
-  return [ mknode "w:hyperlink" [("r:id",ind)] contents ]
+  return [ mknode "w:hyperlink" [("r:id",id')] contents ]
 inlineToOpenXML opts (Image alt (src, tit)) = do
-  exists <- liftIO $ doesFileExist src
-  if exists
-     then do
-       imgs <- gets stImages
-       (ident,size) <- case M.lookup src imgs of
-                            Just (i,img) -> return (i, imageSize img)
-                            Nothing -> do
-                              img <- liftIO $ B.readFile src
-                              let ident' = "image" ++ show (M.size imgs + 1)
-                              let size'  = imageSize img
-                              modify $ \st -> st{
-                                 stImages = M.insert src (ident',img) $ stImages st }
-                              return (ident',size')
-       let (xpt,ypt) = maybe (120,120) sizeInPoints size
-       -- 12700 emu = 1 pt
-       let (xemu,yemu) = (xpt * 12700, ypt * 12700)
-       let cNvPicPr = mknode "pic:cNvPicPr" [] $
-                        mknode "a:picLocks" [("noChangeArrowheads","1"),("noChangeAspect","1")] ()
-       let nvPicPr  = mknode "pic:nvPicPr" []
-                       [ mknode "pic:cNvPr"
-                           [("descr",src),("id","0"),("name","Picture")] ()
-                       , cNvPicPr ]
-       let blipFill = mknode "pic:blipFill" []
-                        [ mknode "a:blip" [("r:embed",ident)] ()
-                        , mknode "a:stretch" [] $ mknode "a:fillRect" [] () ]
-       let xfrm =    mknode "a:xfrm" []
-                       [ mknode "a:off" [("x","0"),("y","0")] ()
-                       , mknode "a:ext" [("cx",show xemu),("cy",show yemu)] () ]
-       let prstGeom = mknode "a:prstGeom" [("prst","rect")] $
-                        mknode "a:avLst" [] ()
-       let ln =      mknode "a:ln" [("w","9525")]
-                       [ mknode "a:noFill" [] ()
-                       , mknode "a:headEnd" [] ()
-                       , mknode "a:tailEnd" [] () ]
-       let spPr =    mknode "pic:spPr" [("bwMode","auto")]
-                       [xfrm, prstGeom, mknode "a:noFill" [] (), ln]
-       let graphic = mknode "a:graphic" [] $
-                       mknode "a:graphicData" [("uri","http://schemas.openxmlformats.org/drawingml/2006/picture")]
-                         [ mknode "pic:pic" []
-                           [ nvPicPr
-                           , blipFill
-                           , spPr ] ]
-       return [ mknode "w:r" [] $
-           mknode "w:drawing" [] $
-             mknode "wp:inline" []
-               [ mknode "wp:extent" [("cx",show xemu),("cy",show yemu)] ()
-               , mknode "wp:effectExtent" [("b","0"),("l","0"),("r","0"),("t","0")] ()
-               , mknode "wp:docPr" [("descr",tit),("id","1"),("name","Picture")] ()
-               , graphic ] ]
-     else do
-       liftIO $ UTF8.hPutStrLn stderr $
-          "Could not find image `" ++ src ++ "', skipping..."
-       inlinesToOpenXML opts alt
+  -- first, check to see if we've already done this image
+  imgs <- gets stImages
+  case M.lookup src imgs of
+    Just (_,_,elt,_) -> return [elt]
+    Nothing -> do
+      let sourceDir = writerSourceDirectory opts
+      res <- liftIO $ E.try $ fetchItem sourceDir src
+      case res of
+        Left (_ :: E.SomeException) -> do
+          liftIO $ warn $ "Could not find image `" ++ src ++ "', skipping..."
+          -- emit alt text
+          inlinesToOpenXML opts alt
+        Right (img, _) -> do
+          ident <- ("rId"++) `fmap` getUniqueId
+          let size = imageSize img
+          let (xpt,ypt) = maybe (120,120) sizeInPoints size
+          -- 12700 emu = 1 pt
+          let (xemu,yemu) = (xpt * 12700, ypt * 12700)
+          let cNvPicPr = mknode "pic:cNvPicPr" [] $
+                           mknode "a:picLocks" [("noChangeArrowheads","1"),("noChangeAspect","1")] ()
+          let nvPicPr  = mknode "pic:nvPicPr" []
+                          [ mknode "pic:cNvPr"
+                              [("descr",src),("id","0"),("name","Picture")] ()
+                          , cNvPicPr ]
+          let blipFill = mknode "pic:blipFill" []
+                           [ mknode "a:blip" [("r:embed",ident)] ()
+                           , mknode "a:stretch" [] $ mknode "a:fillRect" [] () ]
+          let xfrm =    mknode "a:xfrm" []
+                          [ mknode "a:off" [("x","0"),("y","0")] ()
+                          , mknode "a:ext" [("cx",show xemu),("cy",show yemu)] () ]
+          let prstGeom = mknode "a:prstGeom" [("prst","rect")] $
+                           mknode "a:avLst" [] ()
+          let ln =      mknode "a:ln" [("w","9525")]
+                          [ mknode "a:noFill" [] ()
+                          , mknode "a:headEnd" [] ()
+                          , mknode "a:tailEnd" [] () ]
+          let spPr =    mknode "pic:spPr" [("bwMode","auto")]
+                          [xfrm, prstGeom, mknode "a:noFill" [] (), ln]
+          let graphic = mknode "a:graphic" [] $
+                          mknode "a:graphicData" [("uri","http://schemas.openxmlformats.org/drawingml/2006/picture")]
+                            [ mknode "pic:pic" []
+                              [ nvPicPr
+                              , blipFill
+                              , spPr ] ]
+          let imgElt = mknode "w:r" [] $
+               mknode "w:drawing" [] $
+                 mknode "wp:inline" []
+                  [ mknode "wp:extent" [("cx",show xemu),("cy",show yemu)] ()
+                  , mknode "wp:effectExtent" [("b","0"),("l","0"),("r","0"),("t","0")] ()
+                  , mknode "wp:docPr" [("descr",tit),("id","1"),("name","Picture")] ()
+                  , graphic ]
+          modify $ \st -> st{ stImages = M.insert src (ident, imgPath ident img, imgElt, img) $ stImages st }
+          return [imgElt]
+
+imgPath :: String -> B.ByteString -> String
+imgPath ident img = "media/" ++ ident ++
+                            case imageType img of
+                                  Just Png  -> ".png"
+                                  Just Jpeg -> ".jpeg"
+                                  Just Gif  -> ".gif"
+                                  Nothing   -> ""
+
+br :: Element
+br = mknode "w:r" [] [mknode "w:cr" [] () ]
