@@ -32,13 +32,14 @@ Conversion of 'Pandoc' documents to HTML.
 module Text.Pandoc.Writers.HTML ( writeHtml , writeHtmlString ) where
 import Text.Pandoc.Definition
 import Text.Pandoc.Shared
+import Text.Pandoc.Options
 import Text.Pandoc.Templates
 import Text.Pandoc.Generic
 import Text.Pandoc.Readers.TeXMath
 import Text.Pandoc.Slides
 import Text.Pandoc.Highlighting ( highlight, styleToCss,
                                   formatHtmlInline, formatHtmlBlock )
-import Text.Pandoc.XML (stripTags, escapeStringForXML, fromEntities)
+import Text.Pandoc.XML (stripTags, fromEntities)
 import Network.HTTP ( urlEncode )
 import Numeric ( showHex )
 import Data.Char ( ord, toLower )
@@ -52,14 +53,18 @@ import Text.Blaze.Internal(preEscapedString)
 #else
 import Text.Blaze
 #endif
+#if MIN_VERSION_blaze_html(0,5,1)
+import qualified Text.Blaze.XHtml5 as H5
+#else
 import qualified Text.Blaze.Html5 as H5
+#endif
 import qualified Text.Blaze.XHtml1.Transitional as H
 import qualified Text.Blaze.XHtml1.Transitional.Attributes as A
 import Text.Blaze.Renderer.String (renderHtml)
 import Text.TeXMath
 import Text.XML.Light.Output
 import System.FilePath (takeExtension)
-import Data.Monoid (mempty, mconcat)
+import Data.Monoid
 
 data WriterState = WriterState
     { stNotes            :: [Html]  -- ^ List of notes
@@ -76,8 +81,11 @@ defaultWriterState = WriterState {stNotes= [], stMath = False, stQuotes = False,
 -- Helpers to render HTML with the appropriate function.
 
 strToHtml :: String -> Html
-strToHtml = preEscapedString . escapeStringForXML
--- strToHtml = toHtml
+strToHtml ('\'':xs) = preEscapedString "\'" `mappend` strToHtml xs
+strToHtml xs@(_:_)  = case break (=='\'') xs of
+                           (_ ,[]) -> toHtml xs
+                           (ys,zs) -> toHtml ys `mappend` strToHtml zs
+strToHtml [] = ""
 
 -- | Hard linebreak.
 nl :: WriterOptions -> Html
@@ -211,7 +219,10 @@ inTemplate opts tit auths authsMeta date toc body' newvars =
 
 -- | Like Text.XHtml's identifier, but adds the writerIdentifierPrefix
 prefixedId :: WriterOptions -> String -> Attribute
-prefixedId opts s = A.id $ toValue $ writerIdentifierPrefix opts ++ s
+prefixedId opts s =
+  case s of
+    ""       -> mempty
+    _        -> A.id $ toValue $ writerIdentifierPrefix opts ++ s
 
 -- | Replacement for Text.XHtml's unordList.
 unordList :: WriterOptions -> ([Html] -> Html)
@@ -239,8 +250,8 @@ showSecNum = concat . intersperse "." . map show
 -- | Converts an Element to a list item for a table of contents,
 -- retrieving the appropriate identifier from state.
 elementToListItem :: WriterOptions -> Element -> State WriterState (Maybe Html)
-elementToListItem _ (Blk _) = return Nothing
-elementToListItem opts (Sec _ num id' headerText subsecs) = do
+elementToListItem opts (Sec lev num id' headerText subsecs)
+  | lev <= writerTOCDepth opts = do
   let sectnum = if writerNumberSections opts
                    then (H.span ! A.class_ "toc-section-number" $ toHtml $ showSecNum num) >>
                      preEscapedString " "
@@ -250,8 +261,12 @@ elementToListItem opts (Sec _ num id' headerText subsecs) = do
   let subList = if null subHeads
                    then mempty
                    else unordList opts subHeads
-  return $ Just $ (H.a ! A.href (toValue $ "#" ++ writerIdentifierPrefix opts ++ id')
+  return $ Just
+         $ if null id'
+              then (H.a $ toHtml txt) >> subList
+              else (H.a ! A.href (toValue $ "#" ++ writerIdentifierPrefix opts ++ id')
                        $ toHtml txt) >> subList
+elementToListItem _ _ = return Nothing
 
 -- | Convert an Element to Html.
 elementToHtml :: Int -> WriterOptions -> Element -> State WriterState Html
@@ -264,7 +279,7 @@ elementToHtml slideLevel opts (Sec level num id' title' elements) = do
   let titleSlide = slide && level < slideLevel
   header' <- if title' == [Str "\0"]  -- marker for hrule
                 then return mempty
-                else blockToHtml opts (Header level' title')
+                else blockToHtml opts (Header level' (id',[],[]) title')
   let isSec (Sec _ _ _ _ _) = True
       isSec (Blk _)         = False
   innerContents <- mapM (elementToHtml slideLevel opts)
@@ -272,8 +287,9 @@ elementToHtml slideLevel opts (Sec level num id' title' elements) = do
                         -- title slides have no content of their own
                         then filter isSec elements
                         else elements
-  let header'' = if (writerStrictMarkdown opts || writerSectionDivs opts ||
-                     writerSlideVariant opts == S5Slides || slide)
+  let header'' = if (writerSectionDivs opts ||
+                     writerSlideVariant opts == S5Slides ||
+                     slide)
                     then header'
                     else header' ! prefixedId opts id'
   let inNl x = mconcat $ nl opts : intersperse (nl opts) x ++ [nl opts]
@@ -325,7 +341,7 @@ obfuscateLink opts txt s =
               at'      = obfuscateChar '@'
               (linkText, altText) =
                  if txt == drop 7 s' -- autolink
-                    then ("'<code>'+e+'</code>'", name' ++ " at " ++ domain')
+                    then ("e", name' ++ " at " ++ domain')
                     else ("'" ++ txt ++ "'", txt ++ " (" ++ name' ++ " at " ++
                           domain' ++ ")")
           in  case meth of
@@ -376,15 +392,20 @@ treatAsImage fp =
 blockToHtml :: WriterOptions -> Block -> State WriterState Html
 blockToHtml _ Null = return mempty
 blockToHtml opts (Plain lst) = inlineListToHtml opts lst
-blockToHtml opts (Para [Image txt (s,tit)]) = do
+-- title beginning with fig: indicates that the image is a figure
+blockToHtml opts (Para [Image txt (s,'f':'i':'g':':':tit)]) = do
   img <- inlineToHtml opts (Image txt (s,tit))
-  capt <- inlineListToHtml opts txt
+  let tocapt = if writerHtml5 opts
+                  then H5.figcaption
+                  else H.p ! A.class_ "caption"
+  capt <- if null txt
+             then return mempty
+             else tocapt `fmap` inlineListToHtml opts txt
   return $ if writerHtml5 opts
               then H5.figure $ mconcat
-                    [nl opts, img, H5.figcaption capt, nl opts]
+                    [nl opts, img, capt, nl opts]
               else H.div ! A.class_ "figure" $ mconcat
-                    [nl opts, img, H.p ! A.class_ "caption" $ capt,
-                    nl opts]
+                    [nl opts, img, capt, nl opts]
 blockToHtml opts (Para lst) = do
   contents <- inlineListToHtml opts lst
   return $ H.p contents
@@ -392,7 +413,7 @@ blockToHtml _ (RawBlock "html" str) = return $ preEscapedString str
 blockToHtml _ (RawBlock _ _) = return mempty
 blockToHtml opts (HorizontalRule) = return $ if writerHtml5 opts then H5.hr else H.hr
 blockToHtml opts (CodeBlock (id',classes,keyvals) rawCode) = do
-  let tolhs = writerLiterateHaskell opts &&
+  let tolhs = isEnabled Ext_literate_haskell opts &&
                 any (\c -> map toLower c == "haskell") classes &&
                 any (\c -> map toLower c == "literate") classes
       classes' = if tolhs
@@ -427,15 +448,16 @@ blockToHtml opts (BlockQuote blocks) =
      else do
        contents <- blockListToHtml opts blocks
        return $ H.blockquote $ nl opts >> contents >> nl opts
-blockToHtml opts (Header level lst) = do
+blockToHtml opts (Header level (ident,_,_) lst) = do
   contents <- inlineListToHtml opts lst
   secnum <- liftM stSecNum get
   let contents' = if writerNumberSections opts
                      then (H.span ! A.class_ "header-section-number" $ toHtml $ showSecNum secnum) >>
                             strToHtml " " >> contents
                      else contents
-  let contents''  = if writerTableOfContents opts
-                       then H.a ! A.href (toValue $ "#" ++ writerIdentifierPrefix opts ++ "TOC") $ contents'
+  let contents''  = if writerTableOfContents opts && not (null ident)
+                       then H.a ! A.href (toValue $
+                              '#' : writerIdentifierPrefix opts ++ ident) $ contents'
                        else contents'
   return $ (case level of
               1 -> H.h1 contents''
@@ -477,10 +499,13 @@ blockToHtml opts (OrderedList (startnum, numstyle, _) lst) = do
   return $ foldl (!) (ordList opts contents) attribs
 blockToHtml opts (DefinitionList lst) = do
   contents <- mapM (\(term, defs) ->
-                  do term' <- liftM (H.dt) $ inlineListToHtml opts term
+                  do term' <- if null term
+                                 then return mempty
+                                 else liftM (H.dt) $ inlineListToHtml opts term
                      defs' <- mapM ((liftM (\x -> H.dd $ (x >> nl opts))) .
                                     blockListToHtml opts) defs
-                     return $ mconcat $ nl opts : term' : nl opts : defs') lst
+                     return $ mconcat $ nl opts : term' : nl opts :
+                                        intersperse (nl opts) defs') lst
   let lst' = H.dl $ mconcat contents >> nl opts
   let lst'' = if writerIncremental opts
                  then lst' ! A.class_ "incremental"
@@ -576,7 +601,9 @@ inlineToHtml opts inline =
                              Nothing -> return
                                         $ foldl (!) H.code (attrsToHtml opts attr)
                                         $ strToHtml str
-                             Just  h -> return $ foldl (!) h $
+                             Just  h -> do
+                               modify $ \st -> st{ stHighlighting = True }
+                               return $ foldl (!) h $
                                           attrsToHtml opts (id',[],keyvals)
                                 where (id',_,keyvals) = attr
     (Strikeout lst)  -> inlineListToHtml opts lst >>=
@@ -591,7 +618,7 @@ inlineToHtml opts inline =
                                               strToHtml "’")
                               DoubleQuote -> (strToHtml "“",
                                               strToHtml "”")
-                        in  if writerHtml5 opts
+                        in  if writerHtmlQTags opts
                                then do
                                  modify $ \st -> st{ stQuotes = True }
                                  H.q `fmap` inlineListToHtml opts lst
@@ -618,7 +645,7 @@ inlineToHtml opts inline =
                                                 ! A.src (toValue $ url ++ urlEncode str)
                                                 ! A.alt (toValue str)
                                                 ! A.title (toValue str)
-                                  let brtag = if writerHtml5 opts then H5.br else H.br 
+                                  let brtag = if writerHtml5 opts then H5.br else H.br
                                   return $ case t of
                                             InlineMath  -> m
                                             DisplayMath -> brtag >> m >> brtag
@@ -638,7 +665,7 @@ inlineToHtml opts inline =
                                         Left  _ -> inlineListToHtml opts
                                                    (readTeXMath str) >>= return .
                                                      (H.span ! A.class_ "math")
-                               MathJax _ -> return $ toHtml $
+                               MathJax _ -> return $ H.span ! A.class_ "math" $ toHtml $
                                   case t of
                                     InlineMath  -> "\\(" ++ str ++ "\\)"
                                     DisplayMath -> "\\[" ++ str ++ "\\]"
@@ -655,7 +682,9 @@ inlineToHtml opts inline =
                                _             -> return mempty
     (RawInline "html" str) -> return $ preEscapedString str
     (RawInline _ _) -> return mempty
-    (Link [Code _ str] (s,_)) | "mailto:" `isPrefixOf` s ->
+    (Link [Str str] (s,_)) | "mailto:" `isPrefixOf` s &&
+                             s == escapeURI ("mailto" ++ str) ->
+                        -- autolink
                         return $ obfuscateLink opts str s
     (Link txt (s,_)) | "mailto:" `isPrefixOf` s -> do
                         linkText <- inlineListToHtml opts txt
@@ -693,13 +722,21 @@ inlineToHtml opts inline =
                         htmlContents <- blockListToNote opts ref contents
                         -- push contents onto front of notes
                         put $ st {stNotes = (htmlContents:notes)}
-                        return $ H.sup $
-                                 H.a ! A.href (toValue $ "#" ++ writerIdentifierPrefix opts ++ "fn" ++ ref)
-                                     ! A.class_ "footnoteRef"
-                                     ! prefixedId opts ("fnref" ++ ref)
-                                     $ toHtml ref
-    (Cite _ il)  -> do contents <- inlineListToHtml opts il
-                       return $ H.span ! A.class_ "citation" $ contents
+                        let link = H.a ! A.href (toValue $ "#" ++
+                                         writerIdentifierPrefix opts ++ "fn" ++ ref)
+                                       ! A.class_ "footnoteRef"
+                                       ! prefixedId opts ("fnref" ++ ref)
+                                       $ toHtml ref
+                        let link' = case writerEpubVersion opts of
+                                         Just EPUB3 -> link ! customAttribute "epub:type" "noteref"
+                                         _ -> link
+                        return $ H.sup $ link'
+    (Cite cits il)-> do contents <- inlineListToHtml opts il
+                        let citationIds = unwords $ map citationId cits
+                        let result = H.span ! A.class_ "citation" $ contents
+                        return $ if writerHtml5 opts
+                                    then result ! customAttribute "data-cites" (toValue citationIds)
+                                    else result
 
 blockListToNote :: WriterOptions -> String -> [Block] -> State WriterState Html
 blockListToNote opts ref blocks =
@@ -718,4 +755,8 @@ blockListToNote opts ref blocks =
                                   _           -> otherBlocks ++ [lastBlock,
                                                  Plain backlink]
   in  do contents <- blockListToHtml opts blocks'
-         return $ nl opts >> (H.li ! (prefixedId opts ("fn" ++ ref)) $ contents)
+         let noteItem = H.li ! (prefixedId opts ("fn" ++ ref)) $ contents
+         let noteItem' = case writerEpubVersion opts of
+                              Just EPUB3 -> noteItem ! customAttribute "epub:type" "footnote"
+                              _          -> noteItem
+         return $ nl opts >> noteItem'
