@@ -19,25 +19,28 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 {- |
    Module      : Text.Pandoc.Writers.RTF
    Copyright   : Copyright (C) 2006-2010 John MacFarlane
-   License     : GNU GPL, version 2 or above 
+   License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
-   Stability   : alpha 
+   Stability   : alpha
    Portability : portable
 
 Conversion of 'Pandoc' documents to RTF (rich text format).
 -}
-module Text.Pandoc.Writers.RTF ( writeRTF, rtfEmbedImage ) where
+module Text.Pandoc.Writers.RTF ( writeRTF, writeRTFWithEmbeddedImages ) where
 import Text.Pandoc.Definition
+import Text.Pandoc.Options
 import Text.Pandoc.Shared
 import Text.Pandoc.Readers.TeXMath
 import Text.Pandoc.Templates (renderTemplate)
+import Text.Pandoc.Generic (bottomUpM)
 import Data.List ( isSuffixOf, intercalate )
 import Data.Char ( ord, isDigit, toLower )
 import System.FilePath ( takeExtension )
 import qualified Data.ByteString as B
 import Text.Printf ( printf )
 import Network.URI ( isAbsoluteURI, unEscapeString )
+import qualified Control.Exception as E
 
 -- | Convert Image inlines into a raw RTF embedded image, read from a file.
 -- If file not found or filetype not jpeg or png, leave the inline unchanged.
@@ -47,7 +50,8 @@ rtfEmbedImage x@(Image _ (src,_)) = do
   if ext `elem` [".jpg",".jpeg",".png"] && not (isAbsoluteURI src)
      then do
        let src' = unEscapeString src
-       imgdata <- catch (B.readFile src') (\_ -> return B.empty)
+       imgdata <- E.catch (B.readFile src')
+                    (\e -> let _ = (e :: E.SomeException) in return B.empty)
        let bytes = map (printf "%02x") $ B.unpack imgdata
        let filetype = case ext of
                            ".jpg" -> "\\jpegblip"
@@ -61,32 +65,40 @@ rtfEmbedImage x@(Image _ (src,_)) = do
      else return x
 rtfEmbedImage x = return x
 
+-- | Convert Pandoc to a string in rich text format, with
+-- images embedded as encoded binary data.
+writeRTFWithEmbeddedImages :: WriterOptions -> Pandoc -> IO String
+writeRTFWithEmbeddedImages options doc =
+  writeRTF options `fmap` bottomUpM rtfEmbedImage doc
+
 -- | Convert Pandoc to a string in rich text format.
 writeRTF :: WriterOptions -> Pandoc -> String
-writeRTF options (Pandoc (Meta title authors date) blocks) = 
+writeRTF options (Pandoc (Meta title authors date) blocks) =
   let titletext = inlineListToRTF title
       authorstext = map inlineListToRTF authors
       datetext = inlineListToRTF date
       spacer = not $ all null $ titletext : datetext : authorstext
       body = concatMap (blockToRTF 0 AlignDefault) blocks
+      isTOCHeader (Header lev _ _) = lev <= writerTOCDepth options
+      isTOCHeader _ = False
       context = writerVariables options ++
                 [ ("body", body)
                 , ("title", titletext)
                 , ("date", datetext) ] ++
                 [ ("author", a) | a <- authorstext ] ++
                 [ ("spacer", "yes") | spacer ] ++
-                [ ("toc", tableOfContents $ filter isHeaderBlock blocks) |
+                [ ("toc", tableOfContents $ filter isTOCHeader blocks) |
                    writerTableOfContents options ]
   in  if writerStandalone options
          then renderTemplate context $ writerTemplate options
          else body
 
 -- | Construct table of contents from list of header blocks.
-tableOfContents :: [Block] -> String 
+tableOfContents :: [Block] -> String
 tableOfContents headers =
   let contentsTree = hierarchicalize headers
-  in  concatMap (blockToRTF 0 AlignDefault) $ 
-      [Header 1 [Str "Contents"], 
+  in  concatMap (blockToRTF 0 AlignDefault) $
+      [Header 1 nullAttr [Str "Contents"],
        BulletList (map elementToListItem contentsTree)]
 
 elementToListItem :: Element -> [Block]
@@ -100,7 +112,7 @@ elementToListItem (Sec _ _ _ sectext subsecs) = [Plain sectext] ++
 handleUnicode :: String -> String
 handleUnicode [] = []
 handleUnicode (c:cs) =
-  if ord c > 127 
+  if ord c > 127
      then '\\':'u':(show (ord c)) ++ "?" ++ handleUnicode cs
      else c:(handleUnicode cs)
 
@@ -130,32 +142,32 @@ rtfParSpaced :: Int       -- ^ space after (in twips)
              -> Int       -- ^ first line indent (relative to block) (in twips)
              -> Alignment -- ^ alignment
              -> String    -- ^ string with content
-             -> String 
-rtfParSpaced spaceAfter indent firstLineIndent alignment content = 
+             -> String
+rtfParSpaced spaceAfter indent firstLineIndent alignment content =
   let alignString = case alignment of
                            AlignLeft -> "\\ql "
                            AlignRight -> "\\qr "
                            AlignCenter -> "\\qc "
                            AlignDefault -> "\\ql "
   in  "{\\pard " ++ alignString ++
-      "\\f0 \\sa" ++ (show spaceAfter) ++ " \\li" ++ (show indent) ++ 
+      "\\f0 \\sa" ++ (show spaceAfter) ++ " \\li" ++ (show indent) ++
       " \\fi" ++ (show firstLineIndent) ++ " " ++ content ++ "\\par}\n"
 
--- | Default paragraph. 
+-- | Default paragraph.
 rtfPar :: Int       -- ^ block indent (in twips)
        -> Int       -- ^ first line indent (relative to block) (in twips)
        -> Alignment -- ^ alignment
        -> String    -- ^ string with content
-       -> String 
-rtfPar = rtfParSpaced 180 
+       -> String
+rtfPar = rtfParSpaced 180
 
 -- | Compact paragraph (e.g. for compact list items).
 rtfCompact ::  Int       -- ^ block indent (in twips)
            ->  Int       -- ^ first line indent (relative to block) (in twips)
            ->  Alignment -- ^ alignment
            ->  String    -- ^ string with content
-           ->  String 
-rtfCompact = rtfParSpaced 0 
+           ->  String
+rtfCompact = rtfParSpaced 0
 
 -- number of twips to indent
 indentIncrement :: Int
@@ -172,7 +184,7 @@ bulletMarker indent = case indent `mod` 720 of
 
 -- | Returns appropriate (list of) ordered list markers for indent level.
 orderedMarkers :: Int -> ListAttributes -> [String]
-orderedMarkers indent (start, style, delim) = 
+orderedMarkers indent (start, style, delim) =
   if style == DefaultStyle && delim == DefaultDelim
      then case indent `mod` 720 of
               0 -> orderedListMarkers (start, Decimal, Period)
@@ -185,30 +197,30 @@ blockToRTF :: Int       -- ^ indent level
            -> Block     -- ^ block to convert
            -> String
 blockToRTF _ _ Null = ""
-blockToRTF indent alignment (Plain lst) = 
+blockToRTF indent alignment (Plain lst) =
   rtfCompact indent 0 alignment $ inlineListToRTF lst
-blockToRTF indent alignment (Para lst) = 
+blockToRTF indent alignment (Para lst) =
   rtfPar indent 0 alignment $ inlineListToRTF lst
-blockToRTF indent alignment (BlockQuote lst) = 
-  concatMap (blockToRTF (indent + indentIncrement) alignment) lst 
+blockToRTF indent alignment (BlockQuote lst) =
+  concatMap (blockToRTF (indent + indentIncrement) alignment) lst
 blockToRTF indent _ (CodeBlock _ str) =
   rtfPar indent 0 AlignLeft ("\\f1 " ++ (codeStringToRTF str))
 blockToRTF _ _ (RawBlock "rtf" str) = str
 blockToRTF _ _ (RawBlock _ _) = ""
-blockToRTF indent alignment (BulletList lst) = spaceAtEnd $ 
+blockToRTF indent alignment (BulletList lst) = spaceAtEnd $
   concatMap (listItemToRTF alignment indent (bulletMarker indent)) lst
-blockToRTF indent alignment (OrderedList attribs lst) = spaceAtEnd $ concat $ 
+blockToRTF indent alignment (OrderedList attribs lst) = spaceAtEnd $ concat $
   zipWith (listItemToRTF alignment indent) (orderedMarkers indent attribs) lst
-blockToRTF indent alignment (DefinitionList lst) = spaceAtEnd $ 
+blockToRTF indent alignment (DefinitionList lst) = spaceAtEnd $
   concatMap (definitionListItemToRTF alignment indent) lst
-blockToRTF indent _ HorizontalRule = 
+blockToRTF indent _ HorizontalRule =
   rtfPar indent 0 AlignCenter "\\emdash\\emdash\\emdash\\emdash\\emdash"
-blockToRTF indent alignment (Header level lst) = rtfPar indent 0 alignment $
+blockToRTF indent alignment (Header level _ lst) = rtfPar indent 0 alignment $
   "\\b \\fs" ++ (show (40 - (level * 4))) ++ " " ++ inlineListToRTF lst
-blockToRTF indent alignment (Table caption aligns sizes headers rows) = 
+blockToRTF indent alignment (Table caption aligns sizes headers rows) =
   (if all null headers
       then ""
-      else tableRowToRTF True indent aligns sizes headers) ++ 
+      else tableRowToRTF True indent aligns sizes headers) ++
   concatMap (tableRowToRTF False indent aligns sizes) rows ++
   rtfPar indent 0 alignment (inlineListToRTF caption)
 
@@ -230,7 +242,7 @@ tableRowToRTF header indent aligns sizes' cols =
       end = "}\n\\intbl\\row}\n"
   in  start ++ columns ++ end
 
-tableItemToRTF :: Int -> Alignment -> [Block] -> String 
+tableItemToRTF :: Int -> Alignment -> [Block] -> String
 tableItemToRTF indent alignment item =
   let contents = concatMap (blockToRTF indent alignment) item
   in  "{\\intbl " ++ contents ++ "\\cell}\n"
@@ -238,7 +250,7 @@ tableItemToRTF indent alignment item =
 -- | Ensure that there's the same amount of space after compact
 -- lists as after regular lists.
 spaceAtEnd :: String -> String
-spaceAtEnd str = 
+spaceAtEnd str =
   if isSuffixOf "\\par}\n" str
      then (take ((length str) - 6) str) ++ "\\sa180\\par}\n"
      else str
@@ -249,10 +261,10 @@ listItemToRTF :: Alignment  -- ^ alignment
               -> String     -- ^ list start marker
               -> [Block]    -- ^ list item (list of blocks)
               -> [Char]
-listItemToRTF alignment indent marker [] = 
-  rtfCompact (indent + listIncrement) (0 - listIncrement) alignment 
-             (marker ++ "\\tx" ++ (show listIncrement) ++ "\\tab ") 
-listItemToRTF alignment indent marker list = 
+listItemToRTF alignment indent marker [] =
+  rtfCompact (indent + listIncrement) (0 - listIncrement) alignment
+             (marker ++ "\\tx" ++ (show listIncrement) ++ "\\tab ")
+listItemToRTF alignment indent marker list =
   let (first:rest) = map (blockToRTF (indent + listIncrement) alignment) list
       listMarker = "\\fi" ++ show (0 - listIncrement) ++ " " ++ marker ++ "\\tx" ++
                       show listIncrement ++ "\\tab"
@@ -275,7 +287,7 @@ definitionListItemToRTF alignment indent (label, defs) =
   let labelText = blockToRTF indent alignment (Plain label)
       itemsText = concatMap (blockToRTF (indent + listIncrement) alignment) $
                     concat defs
-  in  labelText ++ itemsText 
+  in  labelText ++ itemsText
 
 -- | Convert list of inline items to RTF.
 inlineListToRTF :: [Inline]   -- ^ list of inlines to convert
@@ -291,9 +303,9 @@ inlineToRTF (Strikeout lst) = "{\\strike " ++ (inlineListToRTF lst) ++ "}"
 inlineToRTF (Superscript lst) = "{\\super " ++ (inlineListToRTF lst) ++ "}"
 inlineToRTF (Subscript lst) = "{\\sub " ++ (inlineListToRTF lst) ++ "}"
 inlineToRTF (SmallCaps lst) = "{\\scaps " ++ (inlineListToRTF lst) ++ "}"
-inlineToRTF (Quoted SingleQuote lst) = 
+inlineToRTF (Quoted SingleQuote lst) =
   "\\u8216'" ++ (inlineListToRTF lst) ++ "\\u8217'"
-inlineToRTF (Quoted DoubleQuote lst) = 
+inlineToRTF (Quoted DoubleQuote lst) =
   "\\u8220\"" ++ (inlineListToRTF lst) ++ "\\u8221\""
 inlineToRTF (Code _ str) = "{\\f1 " ++ (codeStringToRTF str) ++ "}"
 inlineToRTF (Str str) = stringToRTF str
@@ -303,11 +315,11 @@ inlineToRTF (RawInline "rtf" str) = str
 inlineToRTF (RawInline _ _) = ""
 inlineToRTF (LineBreak) = "\\line "
 inlineToRTF Space = " "
-inlineToRTF (Link text (src, _)) = 
-  "{\\field{\\*\\fldinst{HYPERLINK \"" ++ (codeStringToRTF src) ++ 
+inlineToRTF (Link text (src, _)) =
+  "{\\field{\\*\\fldinst{HYPERLINK \"" ++ (codeStringToRTF src) ++
   "\"}}{\\fldrslt{\\ul\n" ++ (inlineListToRTF text) ++ "\n}}}\n"
-inlineToRTF (Image _ (source, _)) = 
-  "{\\cf1 [image: " ++ source ++ "]\\cf0}" 
+inlineToRTF (Image _ (source, _)) =
+  "{\\cf1 [image: " ++ source ++ "]\\cf0}"
 inlineToRTF (Note contents) =
-  "{\\super\\chftn}{\\*\\footnote\\chftn\\~\\plain\\pard " ++ 
+  "{\\super\\chftn}{\\*\\footnote\\chftn\\~\\plain\\pard " ++
   (concatMap (blockToRTF 0 AlignDefault) contents) ++ "}"
