@@ -19,7 +19,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 {- |
    Module      : Text.Pandoc.Writers.Textile
    Copyright   : Copyright (C) 2010 John MacFarlane
-   License     : GNU GPL, version 2 or above 
+   License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley.edu>
    Stability   : alpha
@@ -31,7 +31,8 @@ Textile:  <http://thresholdstate.com/articles/4312/the-textile-reference-manual>
 -}
 module Text.Pandoc.Writers.Textile ( writeTextile ) where
 import Text.Pandoc.Definition
-import Text.Pandoc.Shared 
+import Text.Pandoc.Options
+import Text.Pandoc.Shared
 import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.XML ( escapeStringForXML )
 import Data.List ( intercalate )
@@ -46,9 +47,9 @@ data WriterState = WriterState {
 
 -- | Convert Pandoc to Textile.
 writeTextile :: WriterOptions -> Pandoc -> String
-writeTextile opts document = 
-  evalState (pandocToTextile opts document) 
-            (WriterState { stNotes = [], stListLevel = [], stUseTags = False }) 
+writeTextile opts document =
+  evalState (pandocToTextile opts document)
+            (WriterState { stNotes = [], stListLevel = [], stUseTags = False })
 
 -- | Return Textile representation of document.
 pandocToTextile :: WriterOptions -> Pandoc -> State WriterState String
@@ -90,17 +91,18 @@ escapeCharForTextile x = case x of
 escapeStringForTextile :: String -> String
 escapeStringForTextile = concatMap escapeCharForTextile
 
--- | Convert Pandoc block element to Textile. 
+-- | Convert Pandoc block element to Textile.
 blockToTextile :: WriterOptions -- ^ Options
                 -> Block         -- ^ Block element
-                -> State WriterState String 
+                -> State WriterState String
 
 blockToTextile _ Null = return ""
 
-blockToTextile opts (Plain inlines) = 
+blockToTextile opts (Plain inlines) =
   inlineListToTextile opts inlines
 
-blockToTextile opts (Para [Image txt (src,tit)]) = do
+-- title beginning with fig: indicates that the image is a figure
+blockToTextile opts (Para [Image txt (src,'f':'i':'g':':':tit)]) = do
   capt <- blockToTextile opts (Para txt)
   im <- inlineToTextile opts (Image txt (src,tit))
   return $ im ++ "\n" ++ capt
@@ -120,9 +122,10 @@ blockToTextile _ (RawBlock f str) =
 
 blockToTextile _ HorizontalRule = return "<hr />\n"
 
-blockToTextile opts (Header level inlines) = do
+blockToTextile opts (Header level (ident,_,_) inlines) = do
   contents <- inlineListToTextile opts inlines
-  let prefix = 'h' : (show level ++ ". ")
+  let attribs = if null ident then "" else "(#" ++ ident ++ ")"
+  let prefix = 'h' : show level ++ attribs ++ ". "
   return $ prefix ++ contents ++ "\n"
 
 blockToTextile _ (CodeBlock (_,classes,_) str) | any (all isSpace) (lines str) =
@@ -236,7 +239,7 @@ listItemToTextile opts items = do
 
 -- | Convert definition list item (label, list of blocks) to Textile.
 definitionListItemToTextile :: WriterOptions
-                             -> ([Inline],[[Block]]) 
+                             -> ([Inline],[[Block]])
                              -> State WriterState String
 definitionListItemToTextile opts (label, items) = do
   labelText <- inlineListToTextile opts label
@@ -294,8 +297,8 @@ tableRowToTextile opts alignStrings rownum cols' = do
                       0                  -> "header"
                       x | x `rem` 2 == 1 -> "odd"
                       _                  -> "even"
-  cols'' <- sequence $ zipWith 
-            (\alignment item -> tableItemToTextile opts celltype alignment item) 
+  cols'' <- sequence $ zipWith
+            (\alignment item -> tableItemToTextile opts celltype alignment item)
             alignStrings cols'
   return $ "<tr class=\"" ++ rowclass ++ "\">\n" ++ unlines cols'' ++ "</tr>"
 
@@ -320,7 +323,7 @@ tableItemToTextile opts celltype align' item = do
 -- | Convert list of Pandoc block elements to Textile.
 blockListToTextile :: WriterOptions -- ^ Options
                     -> [Block]       -- ^ List of block elements
-                    -> State WriterState String 
+                    -> State WriterState String
 blockListToTextile opts blocks =
   mapM (blockToTextile opts) blocks >>= return . vcat
 
@@ -332,11 +335,11 @@ inlineListToTextile opts lst =
 -- | Convert Pandoc inline element to Textile.
 inlineToTextile :: WriterOptions -> Inline -> State WriterState String
 
-inlineToTextile opts (Emph lst) = do 
+inlineToTextile opts (Emph lst) = do
   contents <- inlineListToTextile opts lst
   return $ if '_' `elem` contents
               then "<em>" ++ contents ++ "</em>"
-              else "_" ++ contents ++ "_" 
+              else "_" ++ contents ++ "_"
 
 inlineToTextile opts (Strong lst) = do
   contents <- inlineListToTextile opts lst
@@ -377,7 +380,7 @@ inlineToTextile opts (Cite _  lst) = inlineListToTextile opts lst
 inlineToTextile _ (Code _ str) =
   return $ if '@' `elem` str
            then "<tt>" ++ escapeStringForXML str ++ "</tt>"
-           else "@" ++ str ++ "@" 
+           else "@" ++ str ++ "@"
 
 inlineToTextile _ (Str str) = return $ escapeStringForTextile str
 
@@ -395,7 +398,10 @@ inlineToTextile _ Space = return " "
 
 inlineToTextile opts (Link txt (src, _)) = do
   label <- case txt of
-                [Code _ s]  -> return s
+                [Code _ s]
+                 | s == src -> return "$"
+                [Str s]
+                 | s == src -> return "$"
                 _           -> inlineListToTextile opts txt
   return $ "\"" ++ label ++ "\":" ++ src
 
